# Standard library imports.
import sys
import os

# Major package imports.
import wx
import numpy as np
import json

# Local imports.
from .tile_manager_base_editor import TileManagerBase
from ..document import DiskImageDocument

from sawx import clipboard
from sawx.filesystem import fsopen
from sawx.utils.processutil import run_detach
from sawx.ui.compactgrid_mouse import DisplayFlags

from atrip.compressor import find_compressors

from .linked_base import LinkedBase
from .. import clipboard_helpers

import logging
log = logging.getLogger(__name__)
clipboard_log = logging.getLogger("clipboard")
event_log = logging.getLogger("event")


class DummyLinkedBase(object):
    segment = None
    segment_uuid = 0

class DummyFocusedViewer(object):
    linked_base = DummyLinkedBase


class ByteEditor(TileManagerBase):
    """Edit binary files

    The views can be restored from data saved in the .omnivore file. The
    metadata file uses the editor name as a keyword into that saved data, so
    multiple editors can save data in the same metadata file without stomping
    over each other.

    The .omnivore file (a JSON file) will contain the editor name as the
    keyword, and editor-specific data below that. Editors should not touch
    anything in the metadata file outside of their keyword, but should save any
    keywords that exist in the file.

    E.g. the file may look like this:

        {
            "omnivore.byte_edit": {
                "layout": {
                    "sidebars": [ ... ],
                    "tile_manager": { .... },
                },
                "viewers": [
                    ...
                ],
                "linked_base_view_segment_uuid" {
                    "uuid1": 0,
                    "uuid2": 3,
                }.
            "omnivore.emulator": {
                ...
            }
        }
    """
    editor_id = "omnivore.byte_edit"
    ui_name = "Byte Editor"

    default_viewers = "hex,bitmap,char,disasm"
    default_viewers = "hex,bitmap,disasm"

    preferences_module = "omnivore.editors.byte_editor_preferences"

    menubar_desc = [
        ["File",
            ["New",
                "new_blank_file",
                None,
                "new_file_from_template",
            ],
            "new_window",
            None,
            "open_file",
            ["Open Recent",
                "open_recent",
            ],
            None,
            "save_file",
            "save_as",
            None,
            "quit",
        ],
        ["Edit",
            "undo",
            "redo",
            None,
            "copy",
            "cut",
            "paste",
            None,
            "select_all",
            "select_hgr",
            "select_none",
            "select_invert",
            ["Mark Selection As",
                "disasm_type",
            ],
            None,
            "find",
            "find_expression",
            "find_next",
            "find_to_selection",
            None,
            "prefs",
        ],
        ["View",
            "view_width",
            "view_zoom",
            ["Colors",
                "view_color_standards",
                None,
                "view_antic_powerup_colors",
                None,
                "view_ask_colors",
            ],
            ["Bitmap",
                "view_bitmap_renderers",
            ],
            ["Text",
                "view_font_renderers",
                None,
                "view_font_mappings",
            ],
            ["Font",
                "view_fonts",
                None,
                "view_font_groups",
                None,
                "view_load_font",
                "view_font_from_selection",
                "view_font_from_segment",
            ],
            None,
            ["Add Data Viewer",
                "view_add_viewer",
            ],
            None,
            "layout_save",
            ["Restore Layout",
                "layout_restore",
            ],
            None,
            "show_toolbar",
        ],
        ["Bytes",
            "byte_set_to_zero",
            "byte_set_to_ff",
            "byte_nop",
            None,
            "byte_set_high_bit",
            "byte_clear_high_bit",
            "byte_bitwise_not",
            "byte_shift_left",
            "byte_shift_right",
            "byte_rotate_left",
            "byte_rotate_right",
            "byte_reverse_bits",
            "byte_random",
            None,
            "byte_set_value",
            "byte_or_with_value",
            "byte_and_with_value",
            "byte_xor_with_value",
            None,
            "byte_ramp_up",
            "byte_ramp_down",
            "byte_add_value",
            "byte_subtract_value",
            "byte_subtract_from",
            "byte_multiply_by",
            "byte_divide_by",
            "byte_divide_from",
            None,
            "byte_reverse_selection",
            "byte_reverse_group",],
        ["Jumpman",
            ["Edit Level",
                "jumpman_level_list",
            ],
            None,
            "jumpman_set_trigger",
            "jumpman_clear_trigger",
            None,
            "jumpman_add_assembly_source",
            "jumpman_compile_assembly_source",
        ],
        ["Generate",
            ["Compression",
                "generate_compression_menu()",
            ],
        ],
        ["Media",
            "generate_segment_menu()",
            None,
            "comment_add",
            "comment_remove",
            None,
            "segment_from_selection",
            "segment_multiple_from_selection",
            "segment_interleave",
            "segment_origin",
            None,
            "segment_goto",
        ],
        ["Machine",
            ["CPU",
                "doc_cpu",
            ],
            ["Operating System",
                "doc_os_labels",
            ],
            ["Assembler Syntax",
                "doc_assembler",
            ],
            None,
            ["Emulator",
                "emu_list",
            ],
            None,
            "emu_boot_skip_frames",
            "emu_boot_start_paused",
            None,
            "emu_boot_disk_image",
            "emu_boot_segment",
            "emu_restore",
        ],
        ["Help",
            "about",
            None,
            "user_guide",
            None,
            "show_debug_log",
            "open_log_directory",
            "open_log",
            None,
            ["Developer Debugging",
                "show_focus",
                "widget_inspector",
                "garbage_objects",
                "raise_exception",
                "test_progress",
            ],
        ],
    ]

    keybinding_desc = {
        "byte_set_to_zero": "Ctrl+0",
        "byte_set_to_ff": "Ctrl+9",
        "byte_nop": "Ctrl+3",
        "byte_set_high_bit": "",
        "byte_clear_high_bit": "",
        "byte_bitwise_not": "Ctrl+1",
        "byte_shift_left": "",
        "byte_shift_right": "",
        "byte_rotate_left": "Ctrl+<",
        "byte_rotate_right": "Ctrl+>",
        "byte_reverse_bits": "",
        "byte_random": "",
        "byte_set_value": "",
        "byte_or_with_value": "Ctrl+\\",
        "byte_and_with_value": "Ctrl+7",
        "byte_xor_with_value": "Ctrl+6",
        "byte_ramp_up": "",
        "byte_ramp_down": "",
        "byte_add_value": "Ctrl+=",
        "byte_subtract_value": "Ctrl+-",
        "byte_subtract_from": "Shift+Ctrl+-",
        "byte_multiply_by": "Ctrl+8",
        "byte_divide_by": "Ctrl+/",
        "byte_divide_from": "Shift+Ctrl+/",
        "byte_reverse_selection": "",
        "byte_reverse_group": "",

        "new_file": "Ctrl+N",
        "open_file": "Ctrl+O",
        "save_file" : "Ctrl+S",
        "save_as" : "Shift+Ctrl+S",
        "cut": "Ctrl+X",
        "copy": "Ctrl+C",
        "paste": "Ctrl+V",
        "undo": "Ctrl+Z",
        "redo": "Shift+Ctrl+Z",
        "select_all": "Ctrl+A",
        "select_none": "Shift+Ctrl+A",
        "select_invert": "Ctrl+I",

        "find": "Ctrl+F",
        "prefs": "Ctrl+,",
        "comment_add": "Ctrl+;",
        "comment_remove": "Shift+Ctrl+;",

        "find_to_selection": "Ctrl+T",  # TEST binding
    }

    module_search_order = ["omnivore.actions", "sawx.actions"]

    # Convenience functions

    @property
    def segment(self):
        return self.focused_viewer.linked_base.segment

    @property
    def segments(self):
        return self.document.segments

    @property
    def linked_base(self):
        return self.focused_viewer.linked_base

    @property
    def segment_uuid(self):
        return self.focused_viewer.linked_base.segment_uuid

    @property
    def section_name(self):
        return str(self.segment)

    @property
    def can_copy(self):
        return self.focused_viewer.can_copy

    @property
    def current_selection(self):
        return self.focused_viewer.linked_base.current_selection

    #### Initialization

    def __init__(self, document, *args, **kwargs):
        TileManagerBase.__init__(self, document, *args, **kwargs)
        self.center_base = None
        self.linked_bases = {}
        self.last_search_settings = {
            "find": "",
            "replace": "",
            "match_case": False,
            "allow_inverse": False,
            "regex": False,
            "algorithm": "",
        }

    @classmethod
    def can_edit_document_exact(cls, document):
        return "atrip_collection" in document.file_metadata

    @classmethod
    def can_edit_document_generic(cls, document):
        return document.mime == "application/octet-stream"

    #### ui

    def generate_segment_menu(self):
        items = []
        for index, container in enumerate(self.document.collection.containers):
            sub_items = [str(container), f"segment_select{{{index}}}"]
            items.append(sub_items)
        return items

    def generate_compression_menu(self):
        items = []
        for index, cls in enumerate(find_compressors()):
            items.append(f"compress_select{{{cls.compression_algorithm}}}")
        items.sort()
        return items

    #### file handling

    def show(self, args=None):
        log.debug(f"show: document {self.document}")
        log.debug(f"show: collection {self.document.collection}")
        log.debug(f"show: segments {self.document.segments}")
        if self.has_command_line_viewer_override(args):
            self.create_layout_from_args(args)
        else:
            s = self.document.last_session.get(self.editor_id, {})
            self.restore_session(s)
        self.restore_legacy_session(self.document.last_session)
        self.set_initial_focused_viewer()
        self.document.recalc_event()

    def create_layout_from_args(self, args):
        log.debug(f"Creating layout from {args}")
        self.center_base = LinkedBase(self)
        self.linked_bases = {self.center_base.uuid:self.center_base}
        viewer_metadata = {}
        for name, value in args.items():
            viewer_metadata[name.strip()] = {}
        self.create_viewers(viewer_metadata)
        self.center_base.view_segment_uuid(None)

    @property
    def layout_template_search_order(self):
        # Try container name, filesystem name
        s = self.center_base.segment
        order = [s.name]
        try:
            order.append(s.filesystem.ui_name)
        except AttributeError:
            pass
        order.append(self.editor_id)
        return order

    def restore_session(self, s):
        log.debug("restore_session: %s" % str(s))
        if 'diff highlight' in s:
            self.diff_highlight = bool(s['diff highlight'])
        self.restore_linked_bases(s)
        self.restore_layout(s)
        # self.restore_view_segment_uuid(s)

    def restore_legacy_session(self, s):
        try:
            container = self.document.collection.containers[0]
        except IndexError:
            pass
        else:
            container.restore_backward_compatible_state(s)

    def restore_linked_bases(self, s):
        linked_bases = {}
        for b in s.get("linked bases", []):
            base = LinkedBase(self)
            base.restore_session(b)
            linked_bases[base.uuid] = base
            log.debug("restore_linked_bases: linked_base[%s]=%s" % (base.uuid, base))
        uuid = s.get("center_base", None)
        log.debug(f"looking for center_base: {uuid}")
        try:
            self.center_base = linked_bases[uuid]
            log.debug(f"found center_base: {self.center_base}")
        except KeyError:
            # no saved session, so find the first interesting segment to display
            segment = self.document.collection.find_interesting_segment_to_edit()
            log.debug(f"most interesting segment: {segment}")
            if segment is None:
                segment = self.document.collection.find_boot_media()
                log.debug(f"default boot media found: {segment}")
            log.debug(f"restoring segment {segment.uuid}")
            self.center_base = LinkedBase(self, segment)
            linked_bases[self.center_base.uuid] = self.center_base

        log.debug(f"linked_bases: {linked_bases}")
        log.debug(f"center_base: {self.center_base}")
        self.linked_bases = linked_bases

    def find_linked_base_by_name(self, segment_name):
        """Check to see if existing linked base is using a segment with the
        specified name
        """
        for linked_base in self.linked_bases.values():
            if linked_base.segment.ui_name == segment_name:
                log.debug(f"find_linked_base_by_name: found {linked_base}")
                return linked_base
        raise KeyError

    def calc_linked_base_by_name(self, segment_name):
        """Find or create a linked base that is using the segment with the
        specified name.
        """
        try:
            linked_base = self.find_linked_base_by_name(segment_name)
        except KeyError:
            for segment in self.document.collection.iter_segments():
                log.debug(f"calc_linked_base_by_name: trying {segment.name}")
                if segment.name == segment_name:
                    log.debug(f"calc_linked_base_by_name: found {segment}")
                    linked_base = LinkedBase(self, segment)
                    break
            else:
                log.debug(f"calc_linked_base_by_name: no segment named {segment_name}")
                raise KeyError(f"No segment with name {segment_name}")
        return linked_base

    def change_initial_segment(self, ui_name):
        """Change the initial viewed segment to the first one that matches
        the ui_name
        """
        for segment in self.document.collection.iter_segments():
            if segment.ui_name == ui_name:
                log.debug(f"change_initial_segment: changing to segment {segment}")
                self.center_base.view_segment_uuid(segment.uuid, False)
                break

    def rebuild_document_properties(self):
        if not self.document.has_baseline:
            self.use_self_as_baseline(self.document)
        FrameworkEditor.rebuild_document_properties(self)
        b = self.focused_viewer.linked_base
        if b.segment_uuid is None:
            self.document.find_initial_visible_segment(b)
        log.debug("rebuilding document %s; initial segment=%s" % (str(self.document), b.segment))
        self.compare_to_baseline()
        self.can_resize_document = self.document.can_resize

    def init_view_properties(self):
        wx.CallAfter(self.force_focus, self.focused_viewer)
        self.task.machine_menu_changed = self.focused_viewer.machine
        # if self.initial_font_segment:
        #     self.focused_viewer.linked_base.machine.change_font_data(self.initial_font_segment)

    def process_preference_change(self, prefs):
        log.debug("%s processing preferences change" % self.task.name)
        #self.machine.set_text_font(prefs.text_font)

    ##### Copy/paste

    @property
    def supported_clipboard_data(self):
        try:
            return self.focused_viewer.supported_clipboard_data
        except AttributeError:
            return []

    def calc_clipboard_data_from(self, focused):
        print("FOCUSED CONTROL", focused)
        # FIXME: for the moment, assume focused control is in focused viewer
        data_objs = self.focused_viewer.control.calc_clipboard_data_objs(focused)
        return data_objs

    def delete_selection_from(self, focused):
        print("FOCUSED CONTROL", focused)
        cmd = self.focused_viewer.calc_delete_command()
        if cmd:
            log.debug(f"deleting selection using command: {cmd}")
            self.process_command(cmd)

    def paste_clipboard(self):
        clipboard_log.debug(f"focused: {self.focused_viewer}")
        data_obj = clipboard.get_clipboard_data(self.supported_clipboard_data)
        if data_obj:
            print("Found data obj", data_obj)
            blob = clipboard_helpers.parse_data_obj(data_obj, self.focused_viewer)
            cmd = self.focused_viewer.calc_paste_command(blob)
            if cmd:
                clipboard_log.debug("processing paste object %s" % cmd)
                self.process_command(cmd)
                return cmd
            # handler = self.get_clipboard_handler(data_obj)
            # if handler:
            #     handler(self, data_obj, focused)
            # else:
            #     clipboard_log.error("No clipboard handler found for data {data_obj}")

    def get_paste_data_from_clipboard(self):
        return clipboard.get_paste_data(self.focused_viewer)

    def process_paste_data(self, serialized_data, cmd_cls=None, *args, **kwargs):
        if cmd_cls is None:
            cmd = self.focused_viewer.get_paste_command(serialized_data, *args, **kwargs)
        else:
            cmd = cmd_cls(self.segment, serialized_data, *args, **kwargs)
        log.debug("processing paste object %s" % cmd)
        self.process_command(cmd)
        return cmd

    #### selection helpers

    def select_all(self):
        self.focused_viewer.select_all()
        self.document.refresh_event(flags=True)

    def select_none(self):
        self.focused_viewer.select_none()
        self.document.refresh_event(flags=True)

    def select_invert(self):
        self.focused_viewer.select_invert()
        self.document.refresh_event(flags=True)

    def check_document_change(self):
        self.document.change_count += 1
        self.update_caret_history()

    def refresh_panes(self):
        log.debug("refresh_panes called")

    def reconfigure_panes(self):
        self.update_pane_names()

    def update_pane_names(self):
        for viewer in self.viewers:
            viewer.update_caption()
        self.control.update_captions()

    def view_segment_uuid(self, uuid):
        base = self.focused_viewer.linked_base
        base.view_segment_uuid(uuid)
        self.update_pane_names()

    def get_extra_segment_savers(self, segment):
        savers = []
        for v in self.viewers:
            savers.extend(v.get_extra_segment_savers(segment))
        return savers

    def save_segment(self, saver, uri):
        try:
            byte_values = saver.encode_data(self.segment, self)
            saver = lambda a,b: byte_values
            self.document.save_to_uri(uri, self, saver, save_metadata=False)
        except Exception as e:
            log.error("%s: %s" % (uri, str(e)))
            #self.window.error("Error trying to save:\n\n%s\n\n%s" % (uri, str(e)), "File Save Error")
            raise

    def show_trace(self):
        """Highlight the current trace after switching to a new segment

        """
        if self.can_trace:
            self.disassembly.update_trace_in_segment()
            self.document.change_count += 1

    ##### Search

    @property
    def search_start(self):
        c = self.focused_viewer.control
        caret = c.caret_handler.current
        index, _ = c.table.get_index_range(caret.rc[0], caret.rc[1])
        return index

    def invalidate_search(self):
        self.task.change_minibuffer_editor(self)

    @property
    def searchers(self):
        search_order = []
        found = set()
        for v in self.viewers:
            for s in v.searchers:
                # searchers may depend on the viewer (like the disassembly)
                # or they may be generic to the segment
                if s.ui_name not in found:
                    search_order.append(s)
                    found.add(s.ui_name)
        log.debug("search order: %s" % [s.ui_name for s in search_order])
        return search_order

    def compare_to_baseline(self):
        if self.diff_highlight and self.document.has_baseline:
            self.document.update_baseline()

    #### command processing

    def calc_status_flags(self):
        return DisplayFlags()

    def process_flags(self, flags):
        """Perform the UI updates described by the DisplayFlags instance
        
        """
        event_log.debug(f"process_flags: {self.ui_name}, flags={flags}")
        d = self.document
        visible_range = False

        #self.caret_handler.process_caret_flags(flags, d)

        if flags.message:
            self.frame.status_message(flags.message)

        if flags.metadata_dirty:
            self.metadata_dirty = True

        control = flags.advance_caret_position_in_control
        if control:
            control.post_process_caret_flags(flags)

        control = flags.sync_caret_from_control
        if control:
            self.linked_base.sync_caret_event(flags=flags)

        if flags.data_model_changed:
            event_log.debug(f"process_flags: data_model_changed")
            d.data_model_changed_event(flags=flags)
            d.change_count += 1
            flags.rebuild_ui = True
        elif flags.byte_values_changed:
            event_log.debug(f"process_flags: byte_values_changed")
            d.change_count += 1
            d.byte_values_changed_event(flags=flags)
            flags.refresh_needed = True
        elif flags.byte_style_changed:
            event_log.debug(f"process_flags: byte_style_changed")
            d.change_count += 1
            d.byte_style_changed_event(flags=flags)
            flags.rebuild_ui = True
            flags.refresh_needed = True

        if flags.rebuild_ui:
            event_log.debug(f"process_flags: rebuild_ui")
            d.recalc_event(flags=flags)
        if flags.refresh_needed:
            event_log.debug(f"process_flags: refresh_needed")
            d.refresh_event(flags=flags)

    #### idle handler

    def idle_when_active(self):
        segment = self.linked_base.segment
        if segment.has_dependent_files:
            segment.check_dependent_files()



# Fast integer multiple scaling of bitmaps

import numpy as np

def intscale(arr, scale, output=None):
    scale = int(scale)
    if scale == 1:
        return arr
    elif scale < 1:
        raise ValueError("Scale must be an integer greater than 1")
    h, w, depth = arr.shape
    if output is None:
        output = np.empty((h * scale, w * scale, depth), dtype=np.uint8)
    dest_h = scale * h
    dest_w = scale * w
    if scale == 2:
        # duplicate source pixels into two columns
        output[::2,::2,:] = arr[:,:,:]
        output[::2,1::2,:] = arr[:,:,:]
        # then copy completed rows to the row below
        output[1::2,:,:] = output[0::2,:,:]
    elif scale == 3:
        # duplicate source pixels into three columns
        output[::3,::3,:] = arr[:,:,:]
        output[::3,1::3,:] = arr[:,:,:]
        output[::3,2::3,:] = arr[:,:,:]
        # then copy completed rows to the next two rows
        output[1::3,:,:] = output[::3,:,:]
        output[2::3,:,:] = output[::3,:,:]
    else:
        raise ValueError("Scale greater than 3 not yet implemented")
        for y in range(h):
            dest_y = y * 3
            # duplicate source pixels to three columns
            output[dest_y,::3,:] = arr[y,::,:]
            output[dest_y,1::3,:] = arr[y,::,:]
            output[dest_y,2::3,:] = arr[y,::,:]
            # then copy that row to two more
            output[dest_y+1,:,:] = output[dest_y,:,:]
            output[dest_y+2,:,:] = output[dest_y,:,:]

    return output



from libc.stdlib cimport malloc, free
import numpy as np
cimport numpy as np

cdef extern:
    int a8bridge_init(int, char **)
    void a8bridge_clear_state_arrays(void *input, void *output)
    void a8bridge_configure_state_arrays(void *input, void *output)
    void a8bridge_get_current_state(void *output)
    void a8bridge_restore_state(void *restore)
    int a8bridge_next_frame(void *input, void *output, void *breakpoints, void *history)
    void a8bridge_show_next_instruction(void *history)

    int libatari800_mount_disk_image(int diskno, const char *filename, int readonly)
    int libatari800_reboot_with_file(const char *filename)


cdef char ** to_cstring_array(list_str):
    cdef char **ret = <char **>malloc(len(list_str) * sizeof(char *))
    for i in xrange(len(list_str)):
        text = list_str[i].encode()
        ret[i] = text
    return ret

def init_emulator(args):
    cdef char *fake_args[10]
    cdef char **argv = fake_args
    cdef int argc
    cdef int err
    cdef char *progname="pyatari800"
    cdef char **c_args = to_cstring_array(args)

    argc = 1
    fake_args[0] = progname
    for i in xrange(len(args)):
        arg = c_args[i]
        fake_args[argc] = arg
        argc += 1

    err = a8bridge_init(argc, argv)
    if err != 1:
        raise RuntimeError(f"Failed starting emulator: error code {err}")
    free(c_args)

def clear_state_arrays(np.ndarray input not None, np.ndarray output not None):
    cdef np.uint8_t[:] ibuf
    cdef np.uint8_t[:] obuf

    ibuf = input.view(np.uint8)
    obuf = output.view(np.uint8)
    a8bridge_clear_state_arrays(&ibuf[0], &obuf[0])

def configure_state_arrays(np.ndarray input not None, np.ndarray output not None):
    cdef np.uint8_t[:] ibuf
    cdef np.uint8_t[:] obuf

    ibuf = input.view(np.uint8)
    obuf = output.view(np.uint8)
    a8bridge_configure_state_arrays(&ibuf[0], &obuf[0])

def next_frame(np.ndarray input not None, np.ndarray output not None, np.ndarray breakpoints not None, history_storage):
    cdef np.uint8_t[:] ibuf
    cdef np.uint8_t[:] obuf
    cdef np.uint8_t[:] dbuf
    cdef np.uint8_t *hbuf
    cdef np.uint8_t[:] tmp
    if history_storage is not None:
        tmp = history_storage.history_array.view(np.uint8)
        hbuf = &tmp[0]
    else:
        hbuf = <np.uint8_t *>0

    ibuf = input.view(np.uint8)
    obuf = output.view(np.uint8)
    dbuf = breakpoints.view(np.uint8)
    bpid = a8bridge_next_frame(&ibuf[0], &obuf[0], &dbuf[0], hbuf)
    return bpid

def show_next_instruction(history_storage):
    cdef np.uint8_t *hbuf
    cdef np.uint8_t[:] tmp
    if history_storage is not None:
        tmp = history_storage.history_array.view(np.uint8)
        hbuf = &tmp[0]
    else:
        hbuf = <np.uint8_t *>0
    a8bridge_show_next_instruction(hbuf)

def get_current_state(np.ndarray output not None):
    cdef np.uint8_t[:] obuf

    obuf = output.view(np.uint8)
    a8bridge_get_current_state(&obuf[0])

def load_disk(int disknum, pathname, int readonly=0):
    filename = pathname.encode('utf-8')
    libatari800_mount_disk_image(disknum, filename, readonly)

def reboot_with_file(pathname):
    filename = pathname.encode('utf-8')
    libatari800_reboot_with_file(filename)

def restore_state(np.ndarray state not None):
    cdef np.uint8_t[:] sbuf
    sbuf = state.view(np.uint8)
    a8bridge_restore_state(&sbuf[0])



import os
import sys
from collections import namedtuple

import wx

from sawx.ui.undo_panel import UndoHistoryPanel
from ..editors.linked_base import VirtualTableLinkedBase
from .segments import SegmentList
from ..viewer import SegmentViewer

import logging
log = logging.getLogger(__name__)



class BaseInfoViewer(SegmentViewer):
    viewer_category = "Info"

    def use_default_view_params(self):
        pass

    def restore_view_params(self, params):
        pass

    def update_toolbar(self):
        pass

    def sync_caret(self, flags):
        pass

    def ensure_visible(self, flags):
        pass

    def recalc_data_model(self):
        self.control.recalc_view()
        self.control.refresh_view()

    def recalc_view(self):
        self.control.recalc_view()
        self.control.refresh_view()

    def get_notification_count(self):
        return 0


class NonCaretInfoViewer(BaseInfoViewer):
    has_caret = False


class VirtualTableInfoViewer(BaseInfoViewer):
    """Info viewer for data that based on data from the segment but doesn't
    display the actual bytes segment in any one-to-one manner. This decouples
    the caret locations so clicking here won't move the cursor in other views.
    """

    @classmethod
    def replace_linked_base(cls, linked_base):
        # the new linked base decouples the cursor here from the other segments
        return VirtualTableLinkedBase(editor=linked_base.editor, segment=linked_base.segment)

    def create_post(self):
        self.linked_base.table = self.control.table

    # override caret display to not respond to caret move requests from other
    # viewers
    def show_caret(self, control, index, bit):
        pass


CommentItem = namedtuple('CommentItem', ('segment', 'index', 'label', 'font', 'type'))

class CommentsPanel(wx.VListBox):
    SEGMENT_TITLE = 0
    COMMENT_ENTRY = 1

    def __init__(self, parent, **kwargs):
        self.items = []
        wx.VListBox.__init__(self, parent, wx.ID_ANY, **kwargs)
        self.SetBackgroundStyle(wx.BG_STYLE_PAINT)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_click)

        f = self.GetFont()
        self.normal_font = f
        self.bold_font = wx.Font(f.GetPointSize(), f.GetFamily(),
                                 f.GetStyle(), wx.BOLD, f.GetUnderlined(),
                                 f.GetFaceName(), f.GetEncoding())
        self.italic_font = wx.Font(f.GetPointSize(), f.GetFamily(),
                                   wx.FONTSTYLE_ITALIC, wx.NORMAL, f.GetUnderlined(),
                                   f.GetFaceName(), f.GetEncoding())
        _, self.font_height = self.GetTextExtent("MWSqj")
        self.select_color = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT)
        self.select_bg_color = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        self.select_brush = wx.Brush(self.select_bg_color, wx.SOLID)
        self.normal_color = self.GetForegroundColour()
        self.SetBackgroundColour(wx.GREEN)

        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)

        # Return key not sent through to EVT_CHAR, EVT_CHAR_HOOK or
        # EVT_KEY_DOWN events in a ListBox. This is the only event handler
        # that catches a Return.
        self.Bind(wx.EVT_KEY_UP, self.on_key_up)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.on_erase_background)

        self.last_segment = None

    def on_erase_background(self, evt):
        """Windows flickers like crazy when erasing the whole screen, so just
        erase the parts that won't be filled in later.
        """
        dc = evt.GetDC()
        w, h = self.GetClientSize()
        print(("erasing background", w, h))
        dc.SetBrush(self.select_brush)
        dc.SetPen(wx.TRANSPARENT_PEN)
        dc.DrawRectangle(0, 0, w, h)

    # This method must be overridden.  When called it should draw the
    # n'th item on the dc within the rect.  How it is drawn, and what
    # is drawn is entirely up to you.
    def OnDrawItem(self, dc, rect, n):
        if self.GetSelection() == n:
            dc.SetTextForeground(self.select_color)
        else:
            dc.SetTextForeground(self.normal_color)
        item = self.items[n]
        dc.SetFont(item.font)
        dc.DrawLabel(item.label, rect,
                     wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL)

    def OnDrawBackground(self, dc, rect, n):
        if self.GetSelection() == n:
            dc.SetBrush(self.select_brush)
        else:
            dc.SetBrush(wx.WHITE_BRUSH)
        dc.SetPen(wx.TRANSPARENT_PEN)
        dc.DrawRectangle(rect)

    # This method must be overridden.  It should return the height
    # required to draw the n'th item.
    def OnMeasureItem(self, n):
        height = 0
        text = self.items[n].label
        for line in text.split('\n'):
            height += self.font_height
        return height + 2

    def DoGetBestSize(self):
        """ Base class virtual method for sizer use to get the best size
        """
        width = 500
        height = -1
        best = wx.Size(width, height)

        # Cache the best size so it doesn't need to be calculated again,
        # at least until some properties of the window change
        self.CacheBestSize(best)

        return best

    def on_key_down(self, evt):
        key = evt.GetKeyCode()
        log.debug("evt=%s, key=%s" % (evt, key))
        moved = False
        index = self.GetSelection()
        if key == wx.WXK_TAB:
            self.process_index(index)
        elif key == wx.WXK_UP:
            index = max(index - 1, 0)
            moved = True
        elif key == wx.WXK_DOWN:
            index = min(index + 1, len(self.items) - 1)
            moved = True
        # elif key == wx.WXK_PAGEUP:
        #     index = self.table.get_page_index(e.caret_index, e.segment.page_size, -1, self)
        #     moved = True
        # elif key == wx.WXK_PAGEDOWN:
        #     index = self.table.get_page_index(e.caret_index, e.segment.page_size, 1, self)
        #     moved = True
        else:
            evt.Skip()
        # if True:
        #     evt.Skip()
        #     return
        if moved:
            self.SetSelection(index)
            self.process_index(index)
            self.Refresh()

    def on_key_up(self, evt):
        keycode = evt.GetKeyCode()
        log.debug("key up: %s" % keycode)
        if keycode == wx.WXK_RETURN:
            pass
        evt.Skip()

    def on_click(self, evt):
        index = self.HitTest(evt.GetPosition())
        if index >= 0:
            self.SetSelection(index)
            self.process_index(index)
            self.Refresh()
        evt.Skip()

    def process_index(self, index):
        v = self.segment_viewer
        item = self.items[index]
        if item.segment == v.segment:
            v.linked_base.sync_caret_to_index(item.index)
        else:
            v.linked_base.find_segment(item.segment)
            v.linked_base.sync_caret_to_index(item.index)
        if item.segment != self.last_segment:
            self.update_items(item.segment)

    def set_items(self, items=None):
        if items is None:
            items = self.input_items
        self.input_items = items
        seen = set()
        self.items = []
        for i in items:
            self.process_comment(i, seen)
        self.SetItemCount(len(self.items))

    def update_items(self, segment):
        self.last_segment = segment
        new_items = []
        font = self.normal_font
        for item in self.items:
            new_item = item
            if item.segment == segment:
                if item.type == self.SEGMENT_TITLE:
                    new_item = item._replace(font=self.bold_font)
                else:
                    new_item = item._replace(font=font)
            else:
                new_item = item._replace(font=self.italic_font)
            new_items.append(new_item)
        self.items = new_items

    def process_comment(self, item, segment_seen):
        v = self.segment_viewer
        segment = v.segment
        try:
            if segment == v.document.segments[0]:
                # if we're currently in the main segment, check for the index
                # in user segments first, rather than listing everything in
                # the main segment.
                raise IndexError
            index = segment.reverse_offset[item[0]]
            if index < 0:
                raise IndexError
            label = segment.get_ui_name_at_index(index)
            font = self.normal_font
            segment_font = self.bold_font
        except IndexError:
            font = self.italic_font
            segment_font = self.italic_font
            # segment, index = v.editor.find_in_user_segment(item[0])
            segment = None
            if segment is not None:
                label = segment.get_ui_name_at_index(index)
            else:
                index = item[0]
                label = "%04x" % index
        show = False
        if segment is None and segment not in segment_seen:
            segment_seen.add(segment)
            show = True
        elif segment is not None and segment.uuid not in segment_seen:
            segment_seen.add(segment.uuid)
            show = True
        if show:
            self.items.append(CommentItem._make((segment, 0, str(segment), segment_font, self.SEGMENT_TITLE)))
        self.items.append(CommentItem._make((segment, index, "  %s: %s" % (label, item[1]), font, self.COMMENT_ENTRY)))

    def recalc_view(self):
        v = self.segment_viewer
        segment = v.segment
        try:
            container = v.segment.container
        except AttributeError:
            comments = []
        else:
            comments = container.get_sorted_comments()
        # print(("COMMENTS!", str(comments)))
        self.set_items(comments)

    def refresh_view(self):
        self.Refresh()


class CommentsViewer(NonCaretInfoViewer):
    name = "comments"

    ui_name = "Comments"

    @classmethod
    def create_control(cls, parent, linked_base, mdict):
        return CommentsPanel(parent, size=(100,500))

    def show_caret(self, control, index, bit):
        pass

    ##### Spring Tab interface

    def get_notification_count(self):
        self.control.recalc_view()
        return len(self.control.items)


class UndoViewer(NonCaretInfoViewer):
    name = "undo"

    ui_name = "Undo History"

    @classmethod
    def create_control(cls, parent, linked_base, mdict):
        return UndoHistoryPanel(parent, linked_base.editor, size=(100,500))

    def show_caret(self, control, index, bit):
        self.control.recalc_view()
        self.control.refresh_view()

    def refresh_view(self, flags):
        self.recalc_view()

    ##### Spring Tab interface

    def get_notification_count(self):
        return 0


class SegmentListViewer(NonCaretInfoViewer):
    name = "segments"

    ui_name = "Segments"

    control_cls = SegmentList

    def recalc_data_model(self):
        pass

    def show_caret(self, control, index, bit):
        pass

    # @on_trait_change('linked_base.editor.document.segments_changed')
    def process_segments_changed(self, evt):
        log.debug("process_segments_changed for %s using %s; flags=%s" % (self.control, self.linked_base, str(evt)))
        if evt is not Undefined:
            self.control.set_segments(self.linked_base.editor.document.segments, self.linked_base.segment_uuid)
            self.recalc_view()

    # # @on_trait_change('linked_base.segment_selected')
    # def process_segment_selected(self, evt):
    #     log.debug("process_segment_selected for %s using %s; flags=%s" % (self.control, self.linked_base, str(evt)))
    #     if evt is not Undefined:
    #         self.control.move_caret(evt)  # evt is segment number

    ##### Spring Tab interface

    def get_notification_count(self):
        return 0



def get_line(start_index, end_index, map_width):
    """Bresenham's Line Algorithm
    Produces a list of tuples from start and end
 
    >>> points1 = get_line((0, 0), (3, 4))
    >>> points2 = get_line((3, 4), (0, 0))
    >>> assert(set(points1) == set(points2))
    >>> print points1
    [(0, 0), (1, 1), (1, 2), (2, 3), (3, 4)]
    >>> print points2
    [(3, 4), (2, 3), (1, 2), (1, 1), (0, 0)]
    
    modified from:
    http://www.roguebasin.com/index.php?title=Bresenham%27s_Line_Algorithm
    """
    # Setup initial conditions
    y1, x1 = divmod(start_index, map_width)
    y2, x2 = divmod(end_index, map_width)
    dx = x2 - x1
    dy = y2 - y1

    # Determine how steep the line is
    is_steep = abs(dy) > abs(dx)

    # Rotate line
    if is_steep:
        x1, y1 = y1, x1
        x2, y2 = y2, x2

    # Swap start and end points if necessary and store swap state
    swapped = False
    if x1 > x2:
        x1, x2 = x2, x1
        y1, y2 = y2, y1
        swapped = True

    # Recalculate differentials
    dx = x2 - x1
    dy = y2 - y1

    # Calculate error
    error = int(dx / 2.0)
    ystep = 1 if y1 < y2 else -1

    # Iterate over bounding box generating points between start and end
    y = y1
    points = []
    for x in range(x1, x2 + 1):
        coord = (y, x) if is_steep else (x, y)
        points.append(coord[1] * map_width + coord[0])
        error -= abs(dy)
        if error < 0:
            y += ystep
            error += dx

    # Reverse the list if the coordinates were swapped
    if swapped:
        points.reverse()
    return points


def get_bounds(start_index, end_index, map_width):
    y1, x1 = divmod(start_index, map_width)
    y2, x2 = divmod(end_index, map_width)

    if x1 > x2:
        x1, x2 = x2, x1
    if y1 > y2:
        y1, y2 = y2, y1
    return (x1, y1), (x2, y2)


def get_rectangle(start_index, end_index, map_width):
    (x1, y1), (x2, y2) = get_bounds(start_index, end_index, map_width)

    points = []
    for x in range(x1, x2 + 1):
        points.append(y1 * map_width + x)
    for y in range(y1 + 1, y2):
        points.append(y * map_width + x1)
        if x2 > x1:
            points.append(y * map_width + x2)
    if y2 > y1:
        for x in range(x1, x2 + 1):
            points.append(y2 * map_width + x)
    return points


def get_filled_rectangle(start_index, end_index, map_width):
    (x1, y1), (x2, y2) = get_bounds(start_index, end_index, map_width)

    points = []
    for y in range(y1, y2 + 1):
        index = y * map_width + x1
        points.extend(list(range(index, index - x1 + x2 + 1)))
    return points



import numpy as np

from .. import errors
from ..media_type import DiskImage

import logging
log = logging.getLogger(__name__)


class Apple16SectorDiskImage(DiskImage):
    ui_name = "Apple ][ Floppy Disk Image (16 sector tracks)"
    sector_size = 256
    expected_size = 143360
    sectors_per_track = 16
    starting_sector_label = 0

    extra_serializable_attributes = ['num_sectors:int', 'first_directory:int', 'max_sectors:int', 'ts_pairs:int', 'dos_release:int', 'last_track_num:int', 'track_alloc_dir:int']

    def init_empty(self):
        super().init_empty()
        self.num_sectors = 0
        self.first_directory = -1
        self.max_sectors = 34 * 16
        self.ts_pairs = -1
        self.dos_release = -1
        self.last_track_num = 34
        self.track_alloc_dir = -1

    def sector_from_track(self, track, sector):
        return (track * self.sectors_per_track) + sector

    def follow_track_sector_pointers(self, track, sector, payload_offset, payload_size):
        """Calculate offsets into the media by following the track/sector
        pointers in the header of each sector, starting at the specified
        track/sector
        """
        offsets = np.empty(self.expected_size, dtype=np.uint32)
        offset_count = 0
        sector = self.sector_from_track(track, sector)
        print(f"starting at sector {sector}")
        while sector > 0:
            pos, count = self.get_index_of_sector(sector)
            offsets[offset_count:offset_count + payload_size] = np.arange(pos + payload_offset, pos + payload_offset + payload_size, dtype=np.int32)
            offset_count += payload_size
            t = self[pos + 1]
            s = self[pos + 2]
            next_sector = self.sector_from_track(t, s)
            print(f"sector: {sector}, pos={pos}, next={next_sector}")
            sector = next_sector
        subset = offsets[0:offset_count].copy()
        return subset

    def follow_track_sector_list(self, ts_segment):
        """Calculate offsets into the media by following the track/sector
        values in the given list of track/sector numbers.

        Each track/sector pair is two bytes in the list.
        """
        offsets = np.empty(self.expected_size, dtype=np.uint32)
        offset_count = 0
        index = 0
        while index < len(ts_segment):
            track = ts_segment[index]
            if track == 0:
                break
            sector = ts_segment[index + 1]
            index += 2
            sector = self.sector_from_track(track, sector)
            pos, count = self.get_index_of_sector(sector)
            print(f"sector: {sector}, pos={pos}, count={count}")
            offsets[offset_count:offset_count + count] = np.arange(pos, pos + count, dtype=np.int32)
            offset_count += count
        subset = offsets[0:offset_count].copy()
        return subset



""" Action definitions for Jumpman Task

"""
import os
import sys

import wx
import numpy as np

from sawx.utils.textutil import text_to_int

from .. import errors
from ..commands import SetRangeValueCommand


import logging
log = logging.getLogger(__name__)


class JumpmanBaseCommand(SetRangeValueCommand):
    def __init__(self, segment, ranges, data, keep_objects_selected=True):
         super().__init__(segment, ranges, data)
         self.keep_objects_selected = keep_objects_selected

    def set_undo_flags(self, flags):
        # FIXME: need to add new flags to control rebuilding of objects, etc?
        flags.byte_values_changed = True

    def do_change(self, editor, undo):
        retval = super().do_change(editor, undo)
        level = self.segment.jumpman_playfield_model
        level.generate_display_objects(self.keep_objects_selected)
        level.cached_screen = None
        log.debug(f"Jumpman do command {self.ui_name}: {level}")
        return retval

    def undo_change(self, editor, undo):
        super().undo_change(editor, undo)
        level = self.segment.jumpman_playfield_model
        level.generate_display_objects(self.keep_objects_selected)
        level.cached_screen = None
        log.debug(f"Jumpman undo command {self.ui_name}: {level}")


class MoveHarvestGridCommand(JumpmanBaseCommand):
    short_name = "move_harvest_grid"
    ui_name = "Move Harvest Grid"

    def __init__(self, segment, dx, dy):
         super().__init__(segment, [(0x46, 0x48)], None)
         self.dx = dx
         self.dy = dy

    def get_data(self, orig):
        return (self.dx, self.dy)


class CreateObjectCommand(JumpmanBaseCommand):
    short_name = "create_jumpman_obj"
    ui_name = "Create Object"


class MoveObjectCommand(JumpmanBaseCommand):
    short_name = "move_jumpman_obj"
    ui_name = "Move Object"


class FlipVerticalCommand(JumpmanBaseCommand):
    short_name = "vflip_jumpman_obj"
    ui_name = "Flip Vertically"


class FlipHorizontalCommand(JumpmanBaseCommand):
    short_name = "hflip_jumpman_obj"
    ui_name = "Flip Horizontally"


class ClearTriggerCommand(JumpmanBaseCommand):
    short_name = "cleartrigger_jumpman_obj"
    ui_name = "Clear Trigger Function"


class SetTriggerCommand(JumpmanBaseCommand):
    short_name = "settrigger_jumpman_obj"
    ui_name = "Set Trigger Function"


class AssemblyChangedCommand(JumpmanBaseCommand):
    short_name = "jumpman_custom_code"
    ui_name = "Reassemble Custom Code"



import numpy as np

from .. import errors
from ..segment import Segment
from ..filesystem import VTOC, Dirent, Directory, Filesystem
from ..file_type import guess_file_type

try:  # Expensive debugging
    _xd = _expensive_debugging
except NameError:
    _xd = False


class AtariDosBootSegment(Segment):
    ui_name = "DOS2 Boot Segment"
    boot_record_type = np.dtype([
        ('BFLAG', 'u1'),
        ('BRCNT', 'u1'),
        ('BLDADR', '<u2'),
        ('BWTARR', '<u2'),
        ])

    extra_serializable_attributes = ['bflag:int', 'brcnt:int', 'bldadr:int']

    def __init__(self, filesystem):
        media = filesystem.media
        indexes, size = self.find_segment_location(media)
        Segment.__init__(self, media, indexes, self.bldadr, name="Boot Sectors", length=size)
        self.segments = self.calc_boot_segments()

    def init_empty(self):
        super().init_empty()
        self.bflag = 0
        self.brcnt = 0
        self.bldadr = 0

    def find_segment_location(self, media):
        try:
            values = media[0:6].view(dtype=self.boot_record_type)[0]
            self.bflag = values['BFLAG']
            if self.bflag == 0:
                # possible boot sector
                self.brcnt = values['BRCNT']
                if self.brcnt == 0:
                    self.brcnt = 3
            else:
                self.brcnt = 3
            self.bldadr = values['BLDADR']
            index, _ = media.get_index_of_sector(1 + self.brcnt)
        except errors.MediaError as e:
            raise errors.IncompatibleMediaError(f"Invalid boot sector: {e}")
        return 0, index

    def calc_boot_segments(self):
        header = Segment(self, 0, self.bldadr, "Boot Header", length=6)
        code = Segment(self, 6, self.bldadr + 6, name="Boot Code", length=len(self) - 6)
        return [header, code]


class AtariDos1SectorVTOC(VTOC):
    ui_name = "DOS2 SD VTOC"

    vtoc_type = np.dtype([
        ('code', 'u1'),
        ('total','<u2'),
        ('unused','<u2'),
        ])

    max_sector = 720

    extra_serializable_attributes = ['total_sectors:int', 'unused_sectors:int']

    def find_segment_location(self):
        media = self.media
        if not media.is_sector_valid(360):
            raise errors.FilesystemError(f"Media ends before sector 360")
        return media.get_contiguous_sectors_offsets(360, 1)

    def calc_sector_map_size(self):
        values = self[0:5].view(dtype=self.vtoc_type)[0]
        code = values[0]
        if code == 0 or code == 2:
            pass
        else:
            raise errors.FilesystemError(f"Invalid VTOC code {code}")
        self.total_sectors = values[1]
        if self.total_sectors > self.max_sector:
            raise errors.FilesystemError(f"Invalid number of sectors {self.total_sectors}")
        self.unused_sectors = values[2]
        return self.max_sector

    def unpack_vtoc(self):
        bits = np.unpackbits(self[0x0a:0x64])
        self.sector_map[0:720] = bits
        if _xd: log.debug("vtoc before:\n%s" % str(self))

    def pack_vtoc(self):
        if _xd: log.debug("vtoc after:\n%s" % str(self))
        packed = np.packbits(self.sector_map[0:720])
        self[0x0a:0x64] = packed


class AtariDos2SectorVTOC(AtariDos1SectorVTOC):
    ui_name = "DOS2 ED VTOC"

    max_sector = 1040

    def find_segment_location(self):
        if self.media.num_sectors != 1040:
            raise errors.FilesystemError(f"Not enhanced density disk")
        return self.media.get_sector_list_offsets([360, 1024]), 0

    def unpack_vtoc(self):
        bits = np.unpackbits(self[0x0a:0x64])
        self.sector_map[0:720] = bits
        bits = np.unpackbits(self[0xd4:0xfa])  # 0x44 - 0x7a in 2nd sector
        self.sector_map[720:1024] = bits
        if _xd: log.debug("vtoc before:\n%s" % str(self))

    def pack_vtoc(self):
        if _xd: log.debug("vtoc after:\n%s" % str(self))
        packed = np.packbits(self.sector_map[0:720])
        self[0x0a:0x64] = packed
        packed = np.packbits(self.sector_map[720:1024])
        self[0xd4:0xfa] = packed


class AtariDosDirent(Dirent):
    ui_name = "DOS2 Dirent"
    extra_serializable_attributes = ['file_num', 'in_use', 'is_sane', 'flag:int', 'num_sectors', 'starting_sector', 'basename', 'ext']

    # ATR Dirent structure described at http://atari.kensclassics.org/dos.htm
    format = np.dtype([
        ('FLAG', 'u1'),
        ('COUNT', '<u2'),
        ('START', '<u2'),
        ('NAME','S8'),
        ('EXT','S3'),
        ])

    FLAG_OPENED_OUTPUT = 0x01
    FLAG_DOS_2 = 0x02
    FLAG_MYDOS = 0x04
    FLAG_IS_DIR = 0x10
    FLAG_LOCKED = 0x20
    FLAG_IN_USE = 0x40
    FLAG_DELETED = 0x80

    def __init__(self, directory, file_num):
        start = file_num * self.format.itemsize
        self.flag = 0  # self.flag needs to have a value, but don't want to got through init_empty twice
        Dirent.__init__(self, directory, file_num, start, self.format.itemsize)

    def init_empty(self):
        super().init_empty()
        self.flag = 0
        self.num_sectors = 0
        self.starting_sector = 0
        self.basename = b''
        self.ext = b''

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.filename == other.filename and self.starting_sector == other.starting_sector and self.num_sectors == other.num_sectors

    @property
    def opened_output(self):
        return bool(self.flag & self.FLAG_OPENED_OUTPUT)

    @property
    def dos_2(self):
        return bool(self.flag & self.FLAG_DOS_2)

    @property
    def mydos(self):
        return bool(self.flag & self.FLAG_MYDOS)

    @property
    def is_dir(self):
        return bool(self.flag & self.FLAG_IS_DIR)

    @property
    def locked(self):
        return bool(self.flag & self.FLAG_LOCKED)

    @property
    def in_use(self):
        return bool(self.flag & self.FLAG_IN_USE)

    @in_use.setter
    def in_use(self, value):
        if value:
            self.flag |= self.FLAG_IN_USE
        else:
            self.flag &= 0xff ^ self.FLAG_IN_USE

    @property
    def deleted(self):
        return bool(self.flag & self.FLAG_DELETED)

    @property
    def filename(self):
        ext = (b'.' + self.ext) if self.ext else b''
        return (self.basename + ext).decode('latin1')

    @property
    def catalog_entry(self):
        return "%03d %-8s%-3s  %03d" % (self.starting_sector, self.basename.decode("latin1"), self.ext.decode("latin1"), self.num_sectors)

    @property
    def status(self):
        output = "o" if self.opened_output else "."
        dos2 = "2" if self.dos_2 else "."
        mydos = "m" if self.mydos else "."
        in_use = "u" if self.in_use else "."
        deleted = "d" if self.deleted else "."
        locked = "*" if self.locked else " "
        flags = "%s%s%s%s%s%s" % (output, dos2, mydos, in_use, deleted, locked)
        return flags

    @property
    def verbose_info(self):
        flags = []
        if self.opened_output: flags.append("OUT")
        if self.dos_2: flags.append("DOS2")
        if self.mydos: flags.append("MYDOS")
        if self.in_use: flags.append("IN_USE")
        if self.deleted: flags.append("DEL")
        if self.locked: flags.append("LOCK")
        return "flags=[%s]" % ", ".join(flags)

    def parse_raw_dirent(self):
        data = self.data[0:16]
        values = data.view(dtype=self.format)[0]
        self.flag = values[0]
        self.num_sectors = int(values[1])
        self.starting_sector = int(values[2])
        self.basename = bytes(values[3]).rstrip()
        self.ext = bytes(values[4]).rstrip()

    def encode_dirent(self):
        data = np.zeros([self.format.itemsize], dtype=np.uint8)
        values = data.view(dtype=self.format)[0]
        values[0] = self.flag
        values[1] = self.num_sectors
        values[2] = self.starting_sector
        values[3] = self.basename
        values[4] = self.ext
        return data

    def get_file(self):
        media = self.filesystem.media
        offsets = np.empty(self.filesystem.max_file_size, dtype=np.uint32)
        length = 0
        next_sector = self.starting_sector
        sectors_remaining = self.num_sectors
        sectors_seen = set()

        while next_sector > 0 and sectors_remaining > 0:
            try:
                index, size = media.get_index_of_sector(next_sector)
            except errors.MediaError:
                length = 0
                self.is_sane = False
                break
            num_bytes = media[index + size - 1]
            file_num = media[index + size - 3] >> 2
            if file_num != self.file_num:
                raise errors.FileNumberMismatchError164(f"Expecting file {self.file_num}, found {file_num}")
            sectors_seen.add(next_sector)

            offsets[length:length + num_bytes] = np.arange(index, index+num_bytes)
            length += num_bytes

            next_sector = ((media[index + size - 3] & 0x3) << 8) + media[index + size - 2]
            if next_sector in sectors_seen:
                raise errors.FileStructureError(f"Bad sector pointer data: attempting to reread sector {next_sector}")
            sectors_remaining -= 1

        if length > 0:
            offsets = np.copy(offsets[0:length])
            file_segment = guess_file_type(media, self.filename, offsets)
            self.segments = [file_segment]
            return file_segment

    def update_sector_info(self, sector_list):
        self.num_sectors = sector_list.num_sectors
        self.starting_sector = sector_list.first_sector

    def add_metadata_sectors(self, vtoc, sector_list, header):
        # no extra sectors are needed for an Atari DOS file; the links to the
        # next sector is contained in the sector.
        pass

    def sanity_check(self):
        media = self.filesystem.media
        if not self.in_use:
            return True
        if not media.is_sector_valid(self.starting_sector):
            return False
        if self.num_sectors < 0 or self.num_sectors > media.num_sectors:
            return False
        return True

    def set_values(self, filename, filetype, index):
        if type(filename) is not bytes:
            filename = filename.encode("latin1")
        if b'.' in filename:
            filename, ext = filename.split(b'.', 1)
        else:
            ext = b'   '
        self.basename = b'%-8s' % filename[0:8]
        self.ext = ext
        self.file_num = index
        self.flag |= self.FLAG_DOS_2 | self.FLAG_IN_USE
        if _xd: log.debug("set_values: %s" % self)


class AtariDos2Directory(Directory):
    ui_name = "DOS2 Directory"

    def find_segment_location(self):
        media = self.media
        if media.is_sector_valid(361):
            if media.sector_size == 256:
                # uses only first 128 bytes
                offsets = media.get_sector_list_offsets(range(361,369), 128)
                return offsets, 0
            else:
                return media.get_contiguous_sectors_offsets(361, 8)
        else:
            raise errors.FilesystemError("Disk image too small to contain a directory")

    def calc_dirents(self):
        segments = []
        for filenum in range(64):
            dirent = AtariDosDirent(self, filenum)
            if dirent.in_use:
                dirent.set_comment_at(0x00, "FILE #%d: Flag" % filenum)
                dirent.set_comment_at(0x01, "FILE #%d: Number of sectors in file" % filenum)
                dirent.set_comment_at(0x03, "FILE #%d: Starting sector number" % filenum)
                dirent.set_comment_at(0x05, "FILE #%d: Filename" % filenum)
                dirent.set_comment_at(0x0d, "FILE #%d: Extension" % filenum)
                segments.append(dirent)
        return segments


class AtariDos2(Filesystem):
    ui_name = "Atari DOS 2"
    default_executable_extension = "XEX"

    def check_media(self, media):
        try:
            media.get_contiguous_sectors
        except AttributeError:
            raise errors.IncompatibleMediaError(f"{self.ui_name} needs sector access")

    def calc_boot_segment(self):
        return AtariDosBootSegment(self)

    def calc_vtoc_segment(self):
        try:
            return AtariDos2SectorVTOC(self)
        except errors.FilesystemError:
            pass
        return AtariDos1SectorVTOC(self)

    def calc_directory_segment(self):
        return AtariDos2Directory(self)



import numpy as np

import wx

from atrip import style_bits

from sawx.utils.permute import bit_reverse_table
from sawx.utils.nputil import intscale, intwscale, intwscale_font

from atrip.machines import atari8bit
from .bitmap_renderers import BaseRenderer
try:
    from . import antic_speedups as speedups
except ImportError:
    speedups = None

import logging
log = logging.getLogger(__name__)


class Mode2(BaseRenderer):
    name = "Antic 2 (Gr 0)"
    char_bit_width = 8
    char_bit_height = 8
    scale_width = 1
    scale_height = 1
    expected_chars = 128

    @classmethod
    def get_image(cls, machine, antic_font, byte_values, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols):
        if speedups is not None:
            array = speedups.get_numpy_font_map_image(machine, antic_font, byte_values, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols)
        else:
            array = get_numpy_font_map_image(machine, antic_font, byte_values, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols)
        return array

    @property
    def bytes_per_char(self):
        return (self.char_bit_width + 7) // 8 * self.char_bit_height

    def get_font(self, data, colors, gr0_colors, reverse=False):
        bits = self.data_to_bits(data)
        font = self.bits_to_font(bits, colors, gr0_colors, reverse)
        return font

    def data_to_bits(self, data):
        """Convert byte data into bits, returning only a subset if
        required
        """
        padded = self.pad_data(data)
        bits = np.unpackbits(padded)
        bits = bits.reshape((-1, 8, 8))
        bits = self.select_subset(bits)
        return bits

    def pad_data(self, data):
        """Expand (or shrink if necessary) to conform to the expected input
        size of font data array
        """
        count = data.shape[0]
        expected_bytes = self.bytes_per_char * self.expected_chars
        if count > expected_bytes:
            data = data[0:expected_bytes]
        elif count < expected_bytes:
            expanded = np.zeros((expected_bytes), dtype=np.uint8)
            expanded[0:count] = data
            expanded[count:] = 0
            data = expanded
        return data

    def select_subset(self, bits):
        """If the font is made up of only a portion of the data, return the
        array that contains only the interesting bits.
        """
        return bits

    def bits_to_font(self, bits, colors, gr0_colors, reverse=False):
        fg, bg = gr0_colors
        if reverse:
            fg, bg = bg, fg
        r = np.empty(bits.shape, dtype=np.uint8)
        r[bits==0] = bg[0]
        r[bits==1] = fg[0]
        g = np.empty(bits.shape, dtype=np.uint8)
        g[bits==0] = bg[1]
        g[bits==1] = fg[1]
        b = np.empty(bits.shape, dtype=np.uint8)
        b[bits==0] = bg[2]
        b[bits==1] = fg[2]
        font = np.zeros((256, 8, 8, 3), dtype=np.uint8)
        # store the first 128 chars as the normal chars
        font[0:128,:,:,0] = r
        font[0:128,:,:,1] = g
        font[0:128,:,:,2] = b

        # create inverse characters from first 128 chars
        r[bits==0] = fg[0]
        r[bits==1] = bg[0]
        g[bits==0] = fg[1]
        g[bits==1] = bg[1]
        b[bits==0] = fg[2]
        b[bits==1] = bg[2]
        font[128:256,:,:,0] = r
        font[128:256,:,:,1] = g
        font[128:256,:,:,2] = b
        return font


class Mode4(Mode2):
    name = "Antic 4 (40x24, 5 color)"

    def bits_to_font(self, bits, colors, gr0_colors, reverse=False):
        """ http://www.atarimagazines.com/compute/issue49/419_1_Graphics_0_Text_In_Four_Colors.php
        
        There are four possible combinations of two bits: 00, 01, 10, 11. Each combination represents a different color. The color corresponding to the bit-pair 00 is stored at location 712; the color for the bit-pair 01 is at location 708; the color for bit-pair 10 is at 709; the color for bit-pair 11 is at 710.
        """
        pf0, pf1, pf2, pf3, bak = colors[4:9]
        r = np.empty(bits.shape, dtype=np.uint8)
        g = np.empty(bits.shape, dtype=np.uint8)
        b = np.empty(bits.shape, dtype=np.uint8)

        c = np.empty((128, 8, 4), dtype=np.uint8)
        c[:,:,0] = bits[:,:,0]*2 + bits[:,:,1]
        c[:,:,1] = bits[:,:,2]*2 + bits[:,:,3]
        c[:,:,2] = bits[:,:,4]*2 + bits[:,:,5]
        c[:,:,3] = bits[:,:,6]*2 + bits[:,:,7]

        pixels = np.empty((128, 8, 4, 3), dtype=np.uint8)
        pixels[c==0] = bak
        pixels[c==1] = pf0
        pixels[c==2] = pf1
        pixels[c==3] = pf2

        font = np.zeros((256, 8, 4, 3), dtype=np.uint8)
        font[0:128,:,:,:] = pixels

        # Inverse characters use pf3 in place of pf2
        pixels[c==3] = pf3
        font[128:256,:,:,:] = pixels

        # create a double-width image to expand the pixels to the correct
        # aspect ratio
        output = intwscale_font(font, 2)
        return output


class Mode5(Mode4):
    name = "Antic 5 (40x12, 5 color)"
    scale_height = 2


class Mode6Base(Mode2):
    name = "Antic 6 (base class)"
    scale_width = 2

    def bits_to_font(self, bits, colors, gr0_colors, reverse=False):
        bg = colors[8]
        r = np.empty(bits.shape, dtype=np.uint8)
        g = np.empty(bits.shape, dtype=np.uint8)
        b = np.empty(bits.shape, dtype=np.uint8)
        font = np.zeros((256, 8, 8, 3), dtype=np.uint8)

        start_char = 0
        for i in range(4, 8):
            end_char = start_char + 64
            fg = colors[i]
            r[bits==0] = bg[0]
            r[bits==1] = fg[0]
            g[bits==0] = bg[1]
            g[bits==1] = fg[1]
            b[bits==0] = bg[2]
            b[bits==1] = fg[2]
            font[start_char:end_char,:,:,0] = r
            font[start_char:end_char,:,:,1] = g
            font[start_char:end_char,:,:,2] = b
            start_char = end_char
        return font


class Mode6Upper(Mode6Base):
    name = "Antic 6 (Gr 1) Uppercase and Numbers"

    def select_subset(self, bits):
        return bits[0:64,:,:]


class Mode6Lower(Mode6Base):
    name = "Antic 6 (Gr 1) Lowercase and Symbols"

    def select_subset(self, bits):
        return bits[64:128,:,:]


class Mode7Upper(Mode6Base):
    name = "Antic 7 (Gr 2) Uppercase and Numbers"
    scale_height = 2

    def select_subset(self, bits):
        return bits[0:64,:,:]


class Mode7Lower(Mode6Base):
    name = "Antic 7 (Gr 2) Lowercase and Symbols"
    scale_height = 2

    def select_subset(self, bits):
        return bits[64:128,:,:]


class Apple2TextMode(Mode2):
    name = "Apple ]["
    char_bit_width = 7
    expected_chars = 256

    def pad_data(self, data):
        """Expand (or shrink if necessary) to conform to the expected input
        size of font data array
        """
        count = data.shape[0]

        # if more than 128 characters, assume not blinking and pad/truncate to
        # 256 chars
        if count >= 2048:
            return data[0:2048]
        if count > 1024:
            expanded = np.zeros(2048, dtype=np.uint8)
            expanded[0:count] = data
            expanded[count:] = 0
            return expanded

        # pad to 1024 if necessary; inverse/blinking will be applied later
        if count < 1024:
            expanded = np.zeros(1024, dtype=np.uint8)
            expanded[0:count] = data
            expanded[count:] = 0
            data = expanded
        return data

    def bits_to_font(self, bits, colors, gr0_colors, reverse=False):
        bg = (0, 0, 0)
        fg = (255, 255, 255)
        r = np.empty(bits.shape, dtype=np.uint8)
        r[bits==0] = bg[0]
        r[bits==1] = fg[0]
        g = np.empty(bits.shape, dtype=np.uint8)
        g[bits==0] = bg[1]
        g[bits==1] = fg[1]
        b = np.empty(bits.shape, dtype=np.uint8)
        b[bits==0] = bg[2]
        b[bits==1] = fg[2]
        font = np.zeros((256, 8, 7, 3), dtype=np.uint8)
        if bits.shape[0] == 256:
            # all 256 chars are defined, so just store them
            font[:,:,:,0] = r[:,:,8:0:-1]
            font[:,:,:,1] = g[:,:,8:0:-1]
            font[:,:,:,2] = b[:,:,8:0:-1]
        else:
            # only 128 chars are present, so create inversed/blink copies

            # Normal characters get stored in 2nd 128 char positions
            font[128:256,:,:,0] = r[:,:,8:0:-1]
            font[128:256,:,:,1] = g[:,:,8:0:-1]
            font[128:256,:,:,2] = b[:,:,8:0:-1]

            # First 64 are inversed
            r[bits==0] = fg[0]
            r[bits==1] = bg[0]
            g[bits==0] = fg[1]
            g[bits==1] = bg[1]
            b[bits==0] = fg[2]
            b[bits==1] = bg[2]
            font[0:64,:,:,0] = r[0:64,:,8:0:-1]
            font[0:64,:,:,1] = g[0:64,:,8:0:-1]
            font[0:64,:,:,2] = b[0:64,:,8:0:-1]

            # Next 64 are blinking!
            if reverse:
                fg, bg = bg, fg
            r[bits==0] = fg[0]
            r[bits==1] = bg[0]
            g[bits==0] = fg[1]
            g[bits==1] = bg[1]
            b[bits==0] = fg[2]
            b[bits==1] = bg[2]
            font[64:128,:,:,0] = r[0:64,:,8:0:-1]
            font[64:128,:,:,1] = g[0:64,:,8:0:-1]
            font[64:128,:,:,2] = b[0:64,:,8:0:-1]
        return font


def get_numpy_font_map_image(segment_viewer, antic_font, byte_values, style, start_byte, end_byte, bytes_per_row, num_rows, start_col, num_cols):
    width = int(antic_font.char_w * num_cols)
    height = int(num_rows * antic_font.char_h)
    log.debug("pixel width: %dx%d" % (width, height))
    array = np.empty((height, width, 3), dtype=np.uint8)

    log.debug("start byte: %s, end_byte: %s, bytes_per_row=%d num_rows=%d start_col=%d num_cols=%d" % (start_byte, end_byte, bytes_per_row, num_rows, start_col, num_cols))
    end_col = min(bytes_per_row, start_col + num_cols)
    y = 0
    e = start_byte
    f = antic_font.normal_font
    fh = antic_font.highlight_font
    fd = antic_font.data_font
    fm = antic_font.match_font
    fc = antic_font.comment_font
    char_w = antic_font.char_w
    char_h = antic_font.char_h
    mapping = segment_viewer.font_mapping.font_mapping
    for j in range(num_rows):
        x = 0
        for i in range(start_col, start_col + num_cols):
            if e + i >= end_byte or i >= end_col:
                array[y:y+char_h,x:x+char_w,:] = segment_viewer.preferences.background_color
            else:
                c = mapping[byte_values[j, i]]
                s = style[j, i]
                if s & style_bits.selected_bit_mask:
                    array[y:y+char_h,x:x+char_w,:] = fh[c]
                elif s & style_bits.match_bit_mask:
                    array[y:y+char_h,x:x+char_w,:] = fm[c]
                elif s & style_bits.comment_bit_mask:
                    array[y:y+char_h,x:x+char_w,:] = fc[c]
                elif s & style_bits.user_bit_mask:
                    array[y:y+char_h,x:x+char_w,:] = fd[c]
                else:
                    array[y:y+char_h,x:x+char_w,:] = f[c]
            x += char_w
        y += char_h
        e += bytes_per_row
    return array



font_renderer_list = [
    Mode2(),
    Mode4(),
    Mode5(),
    Mode6Upper(),
    Mode6Lower(),
    Mode7Upper(),
    Mode7Lower(),
    Apple2TextMode(),
]

valid_font_renderers = {item.name: item for item in font_renderer_list}



"""Segment actions
"""
import os
import sys

import wx

from sawx.ui.dialogs import prompt_for_string

from ..action import ViewerAction

from ..commands import comment
from .. import errors

import logging
log = logging.getLogger(__name__)


def prompt_for_comment(e, s, ranges, desc):
    existing = s.get_first_comment(ranges)
    text = prompt_for_string(e.control, desc, "Add Comment", existing)
    if text is not None:
        cmd = comment.SetCommentCommand(s, ranges, text)
        e.process_command(cmd)


class comment_add(ViewerAction):
    """Add a text comment to a byte location.

    A comment is associated with a single byte, so although a range can be
    selected, the comment is applied to only the first byte in the range.

    Bytes with comments will be highlighted in all displays.
    """
    name = 'Add Comment'

    def calc_enabled(self, action_key):
        return True

    def is_range(self):
        ch = self.viewer.control.caret_handler
        return ch.has_selection or len(ch) > 1

    def perform(self, action_key):
        e = self.editor
        s = e.segment
        if self.is_range():
            ranges = self.viewer.control.get_selected_ranges_including_carets()
            if len(ranges) == 1:
                desc = "Enter comment for first byte of range:\n%s" % self.viewer.get_label_of_first_byte(ranges)
            else:
                desc = "Enter comment for first byte of each range:\n%s" % self.viewer.get_label_of_first_byte(ranges)
        else:
            ranges = []
        if not ranges:
            index = self.viewer.control.get_current_caret_index()
            ranges = [(index, index+1)]
            desc = "Enter comment for location %s" % index
        prompt_for_comment(e, s, ranges, desc)


class comment_remove(ViewerAction):
    """Remove any comments that are in the selected range, or if no selection
    from the current caret position.

    """
    name = 'Remove Comment'
    accelerator = 'Shift+Alt+C'

    def perform(self, event):
        e = self.editor
        s = e.segment
        ranges = self.viewer.control.get_selected_ranges_including_carets()
        if ranges:
            cmd = comment.ClearCommentCommand(s, ranges)
            e.process_command(cmd)



import ctypes
import time
import numpy as np

from . import libatari800 as liba8
from . import dtypes as d
from . import akey
from .colors import NTSC

try:
    from ..ui.screen import BitmapScreen, OpenGLScreen, GLSLScreen
except ImportError:
    pass



import os

import numpy as np
import pytest

from atrip.assemblers import mac65

class TestMac65:
    def setup(self):
        self.mac65 = mac65.MAC65(True)

    @pytest.mark.parametrize(("pathname"), ["../README.rst", "test_data/fails.m65"])
    def test_failures(self, pathname):
        asm = self.mac65.assemble(pathname)
        assert not asm

    def test_lasers(self):
        asm = self.mac65.assemble("../samples/lasers.s")
        assert asm
        assert "pewpew" in asm.labels

    def test_works(self):
        asm = self.mac65.assemble("test_data/works.m65")
        assert asm
        assert "begin" in asm.labels

    def test_parse_lst(self):
        c = self.mac65
        result = mac65.AssemblerResult()
        with open("../samples/lasers.s.lst") as fh:
            text = fh.read()
        c.verbose = True
        c.parse(result, text)
        print(result.labels)
        print(result.addr_to_label)


if __name__ == "__main__":
    t = TestMac65()
    t.setup()
    t.test_parse_lst()



import hashlib
import inspect
import pkg_resources

import numpy as np

from . import errors
from . import style_bits
from .segment import Segment
from .utils import to_numpy, to_numpy_list, uuid
from .file_type import guess_file_type

import logging
log = logging.getLogger(__name__)

try:  # Expensive debugging
    _xd = _expensive_debugging
except NameError:
    _xd = False


class Filesystem:
    """Base class for a "filesystem", which takes a source segment and
    subdivides it into a set of segments where each represents a file. Some
    auxiliary segments include a `VTOC` and a list of `Dirent`s that point to
    files.

    It is used to fill a `Media` instance with segments.
    """
    ui_name = "Filesystem"

    extra_serializable_attributes = []

    def __init__(self, media):
        self.check_media(media)
        self.media = media
        self.boot = self.calc_boot_segment()
        self.vtoc = self.calc_vtoc_segment()
        self.directory = self.calc_directory_segment()
        media.segments = list(self.filesystem_metadata_segments())

    #### initialization

    def check_media(self, media):
        """Subclasses should override this method to verify the media type is
        supported by the filesystem.

        Subclasses should raise IncompatibleMediaError if the filesystem is not
        possible on this media, for instance attempting to use a disk
        filesystem on a cassette media image.
        """
        pass

    def calc_boot_segment(self):
        """Subclasses should override this method to create a boot segment if
        the filesystem supports one and it is present.

        If it is present, return a single `Segment` instance comprising the
        entire set of data, and use sub-segments if more detail is present.

        If this feature is not present, return None.

        Subclasses should raise the appropriate FilesystemError if the data is
        incompatible with this filesystem.
        """
        pass

    def calc_vtoc_segment(self):
        """Subclasses should override this method to create a VTOC segment if
        the filesystem supports one and it is present.

        If it is present, return a single `Segment` instance comprising the
        entire set of data, and use sub-segments if more detail is present.

        If this feature is not present, return None.

        Subclasses should raise the appropriate FilesystemError if the data is
        incompatible with this filesystem.
        """
        pass

    def calc_directory_segment(self):
        """Subclasses should override this method to create a directory segment
        if the filesystem supports one and it is present.

        If it is present, return a single `Segment` instance comprising the
        entire set of data, and use sub-segments if more detail is present.

        If this feature is not present, return None.

        Subclasses should raise the appropriate FilesystemError if the data is
        incompatible with this filesystem.
        """
        pass

    ####

    @property
    def max_file_size(self):
        return len(self.media)

    #### iterators

    def filesystem_metadata_segments(self):
        if self.boot is not None:
            yield self.boot
        if self.vtoc is not None:
            yield self.vtoc
        if self.directory is not None:
            yield self.directory

    def iter_dirents(self):
        for segment in self.media.segments:
            if isinstance(segment, Dirent):
                yield segment
            yield from segment.iter_segments(Dirent)

    #### utilities

    def find_interesting_segment(self):
        """If the filesystem has some interesting segment that might be better
        to display than the first segment in the container, identify that here
        and an editor can use this info to display that segment in the initial
        view.
        """
        return None


class NoFilesystem(Filesystem):
    ui_name = "No Filesystem"

    def filesystem_metadata_segments(self):
        media = self.media
        segment = guess_file_type(media, media.container.basename, 0, len(media))
        yield segment


class Dirent(Segment):
    """Abstract base class for a directory entry

    """
    ui_name = "Dirent"
    extra_serializable_attributes = ['file_num', 'in_use', 'is_sane']

    def __init__(self, directory, file_num, start, length, parent=None):
        self.directory = directory
        self.file_num = file_num
        self.in_use = True
        self.is_sane = True
        if parent is None:
            parent = directory
        Segment.__init__(self, parent, start, name=f"{self.ui_name} {file_num}", length=length)
        self.init_dirent()

    def init_dirent(self):
        self.parse_raw_dirent()
        if self.sanity_check():
            self.is_sane = True
        else:
            self.is_sane = False
            self.in_use = False
        if self.in_use:
            try:
                self.get_file()
            except errors.FileError as e:
                self.in_use = False
                self.is_sane = False
                self.error = e

    def __eq__(self, other):
        raise NotImplementedError

    def __str__(self):
        return "File #%-2d %s %s" % (self.file_num, f"({self.status})" if self.status else "", self.catalog_entry)

    @property
    def verbose_info(self):
        return ""

    @property
    def status(self):
        return ""

    @property
    def catalog_entry(self):
        return self.filename

    @property
    def filesystem(self):
        return self.directory.filesystem

    @property
    def media(self):
        return self.directory.filesystem.media

    def sanity_check(self):
        return True

    def extra_metadata(self, image):
        return self.verbose_info

    def mark_deleted(self):
        self.deleted = True
        self.in_use = False

    def parse_raw_dirent(self):
        raise NotImplementedError

    def encode_dirent(self):
        raise NotImplementedError

    def get_file(self):
        raise NotImplementedError


class Directory(Segment):
    ui_name = "Directory"

    def __init__(self, filesystem):
        self.filesystem = filesystem
        offset, length = self.find_segment_location()
        Segment.__init__(self, filesystem.media, offset, name=self.ui_name, length=length)

        # Each segment is a dirent
        self.segments = self.calc_dirents()

    @property
    def media(self):
        return self.filesystem.media

    def find_segment_location(self):
        raise NotImplementedError("Subclasses must define this to declare where the directory segment is located in the media image")

    def calc_dirents(self):
        raise NotImplementedError("Subclasses must define this to generate a list of Dirent segments")

    def set(self, index, dirent):
        self.segments[index] = dirent
        if _xd: log.debug("set dirent #%d: %s" % (index, dirent))

    def get_free_dirent(self):
        used = set()
        d = list(self.segments.items())
        if d:
            d.sort()
            for i, dirent in d:
                if not dirent.in_use:
                    return i
                used.add(i)
            if self.num_dirents > 0 and (len(used) >= self.num_dirents):
                raise errors.NoSpaceInDirectory()
            i += 1
        else:
            i = 0
        used.add(i)
        return i

    def add_dirent(self, filename, filetype):
        index = self.get_free_dirent()
        dirent = self.dirent_class(None)
        dirent.set_values(filename, filetype, index)
        self.set(index, dirent)
        return dirent

    def find_dirent(self, filename):
        if hasattr(filename, "filename"):
            # we've been passed a dirent instead of a filename
            for dirent in list(self.segments.values()):
                if dirent == filename:
                    return dirent
        else:
            for dirent in list(self.segments.values()):
                if filename == dirent.filename:
                    return dirent
        raise errors.FileNotFound("%s not found on disk" % filename)

    def save_dirent(self, image, dirent, vtoc, sector_list):
        vtoc.assign_sector_numbers(dirent, sector_list)
        dirent.add_metadata_sectors(vtoc, sector_list, image.header)
        dirent.update_sector_info(sector_list)
        self.calc_sectors(image)

    def remove_dirent(self, image, dirent, vtoc, sector_list):
        vtoc.free_sector_list(sector_list)
        dirent.mark_deleted()
        self.calc_sectors(image)

    @property
    def dirent_class(self):
        raise NotImplementedError

    def calc_sectors(self, image):
        self.sectors = []
        self.current_sector = self.get_dirent_sector()
        self.encode_index = 0

        d = list(self.segments.items())
        d.sort()
        # there may be gaps, so fill in missing entries with blanks
        current = 0
        for index, dirent in d:
            for missing in range(current, index):
                if _xd: log.debug("Encoding empty dirent at %d" % missing)
                data = self.encode_empty()
                self.store_encoded(data)
            if _xd: log.debug("Encoding dirent: %s" % dirent)
            data = self.encode_dirent(dirent)
            self.store_encoded(data)
            current = index + 1
        self.finish_encoding(image)

    def get_dirent_sector(self):
        return self.sector_class(self.sector_size)

    def encode_empty(self):
        raise NotImplementedError

    def encode_dirent(self, dirent):
        raise NotImplementedError

    def store_encoded(self, data):
        while True:
            if _xd: log.debug("store_encoded: %d bytes in %s" % (len(data), self.current_sector))
            data = self.current_sector.add_data(data)
            if len(data) > 0:
                self.sectors.append(self.current_sector)
                self.current_sector = self.get_dirent_sector()
            else:
                break

    def finish_encoding(self, image):
        if not self.current_sector.is_empty:
            self.sectors.append(self.current_sector)
        self.set_sector_numbers(image)

    def set_sector_numbers(self, image):
        raise NotImplementedError


class VTOC(Segment):
    ui_name = "VTOC"

    extra_serializable_attributes = ['addressable_sectors', 'sector_map']

    def __init__(self, filesystem):
        self.filesystem = filesystem
        offset, length = self.find_segment_location()
        Segment.__init__(self, filesystem.media, offset, name=self.ui_name, length=length)
        self.addressable_sectors = self.calc_sector_map_size()
        self.create_sector_map()
        self.unpack_vtoc()

    @property
    def media(self):
        return self.filesystem.media

    def find_segment_location(self):
        """Calculate the location on the media for the VTOC. Return either
        sector number and count, or offset list
        """
        raise NotImplementedError("Subclasses must define this to declare where the directory segment is located in the media image")

    # def __str__(self):
    #     return "%s\n (%d free)" % ("\n".join(["track %02d: %s" % (i, line) for i, line in enumerate(str(self.sector_map[self.header.starting_sector_label:(self.header.tracks_per_disk*self.header.sectors_per_track) + self.header.starting_sector_label].reshape([self.header.tracks_per_disk,self.header.sectors_per_track])).splitlines())]), self.num_free_sectors)

    def calc_sector_map_size(self):
        return self.media.num_sectors

    def create_sector_map(self):
        # sector map: 1 is free, 0 is allocated
        self.sector_map = np.zeros([self.addressable_sectors], dtype=np.uint8)

    def unpack_vtoc(self):
        """Using the bit-encoded data, unpack it into the sector_map array
        """
        raise NotImplementedError

    def pack_vtoc(self):
        """Pack the sector_map array into the segment
        """
        raise NotImplementedError

    @property
    def num_free_sectors(self):
        free = np.where(self.sector_map == 1)[0]
        return len(free)

    def iter_free_sectors(self):
        for i, pos, size in self.filesystem.media.iter_sectors():
            if self.sector_map[i] == 1:
                yield i, pos, size

    def assign_sector_numbers(self, dirent, sector_list):
        """ Map out the sectors and link the sectors together

        raises NotEnoughSpaceOnDisk if the whole file won't fit. It will not
        allow partial writes.
        """
        num = len(sector_list)
        order = self.reserve_space(num)
        if len(order) != num:
            raise errors.NotEnoughSpaceOnDisk(f"Need {num} sectors, VTOC has only {len(order)} available")
        file_length = 0
        last_sector = None
        for sector, sector_num in zip(sector_list.sectors, order):
            sector.sector_num = sector_num
            sector.file_num = dirent.file_num
            file_length += sector.used
            if last_sector is not None:
                last_sector.next_sector_num = sector_num
            last_sector = sector
        if last_sector is not None:
            last_sector.next_sector_num = 0
        sector_list.file_length = file_length

    def reserve_space(self, num):
        order = []
        for i in range(num):
            order.append(self.get_next_free_sector())
        if _xd: log.debug("Sectors reserved: %s" % order)
        self.calc_bitmap()
        return order

    def get_next_free_sector(self):
        free = np.nonzero(self.sector_map)[0]
        if len(free) > 0:
            num = free[0]
            if _xd: log.debug("Found sector %d free" % num)
            self.sector_map[num] = 0
            return num
        raise errors.NotEnoughSpaceOnDisk("No space left in VTOC")

    def calc_bitmap(self):
        raise NotImplementedError

    def free_sector_list(self, sector_list):
        for sector in sector_list:
            self.sector_map[sector.sector_num] = 1
        self.calc_bitmap()


_filesystems = None

def _find_filesystems():
    filesystems = []
    for entry_point in pkg_resources.iter_entry_points('atrip.filesystems'):
        mod = entry_point.load()
        log.debug(f"find_filesystems: Found module {entry_point.name}={mod.__name__}")
        for name, obj in inspect.getmembers(mod):
            if inspect.isclass(obj) and Filesystem in obj.__mro__[1:]:
                log.debug(f"find_filesystems:   found media_type class {name}")
                filesystems.append(obj)
    return filesystems

def find_filesystems():
    global _filesystems

    if _filesystems is None:
        _filesystems = _find_filesystems()
    return _filesystems

def guess_filesystem(segment):
    for f in find_filesystems():
        log.debug(f"trying filesystem {f.ui_name}")
        try:
            found = f(segment)
        except errors.FilesystemError as e:
            log.debug(f"found error: {e}")
            continue
        except errors.MediaError as e:
            log.debug(f"found media error: {e}")
            continue
        else:
            log.info(f"found filesystem {f.ui_name}")
            return found
    log.info(f"No recognized filesystem.")
    return NoFilesystem(segment)



import gzip
import io
import zipfile

import numpy as np

from .. import errors
from ..archiver import Archiver

import logging
log = logging.getLogger(__name__)


class ZipArchiver(Archiver):
    archive_type = "zip"

    def iter_archive(self, basename, byte_data):
        byte_data = io.BytesIO(byte_data)
        try:
            with zipfile.ZipFile(byte_data) as zf:
                for item in zf.infolist():
                    log.debug(f"zipinfo item: {item}")
                    item_data = np.frombuffer(zf.open(item).read(), dtype=np.uint8)
                    yield item.filename, item_data
        except:
            raise errors.InvalidArchiver("Not a zip file")

    def pack_data(self, fh, containers, skip_missing_compressors=False):
        with zipfile.ZipFile(fh, 'w', zipfile.ZIP_DEFLATED, False) as zf:
            for c in containers:
                byte_data = c.calc_packed_bytes(skip_missing_compressors)
                zf.writestr(c.pathname, byte_data)



cdef extern:
    int py_assemble(char *, char *, char *)

def mac65_assemble(source):
    cdef char *source_c
    cdef char *listfile_c
    
    source_c = source
    listfile = source + b".lst"
    listfile_c = listfile
    errfile = source + b".err"
    errfile_c = errfile

    exitval = py_assemble(source_c, listfile_c, errfile_c)
    try:
        with open(errfile, "r") as fh:
            errors = fh.read()
        #print errors
    except IOError as e:
        return str(e), None
    if exitval == 1:
        #print "ERROR!!!!"
        text = None
    else:
        #print "source", source
        #print "opening -->", listfile, "<--"
        with open(listfile, "r") as fh:
            text = fh.read()
            #print text

    #print "exitval=", exitval
    return errors, text



import textwrap

import numpy as np

from .. import errors
from ..stringifier import Stringifier


class CBytes(Stringifier):
    text_type = "c_bytes"
    ui_name = "C bytes"

    def calc_text(self, byte_data):
        text = ",\n".join([",".join(["0x%02x" % d for d in c]) for c in [byte_data[i:i+8] for i in range(0, len(byte_data), 8)]])
        return text



import os
import sys
import pickle

import numpy as np
import wx
import wx.lib.filebrowsebutton as filebrowse
from wx.lib.expando import ExpandoTextCtrl, EVT_ETC_LAYOUT_NEEDED

from atrip import get_xex

from sawx.utils.processutil import which
from sawx.utils.textutil import text_to_int
from sawx.ui.dropscroller import ReorderableList, PickledDropTarget, PickledDataObject
from sawx.ui.dialogs import DictEditDialog

from ..document import DiskImageDocument


class AssemblerDialog(DictEditDialog):
    border = 5

    def __init__(self, parent, title, default=None):
        fields = [
            ('text', 'name', 'Name: '),
            ('text', 'origin', 'Origin Directive: '),
            ('text', 'data byte', 'Data Byte Directive: '),
            ('text', 'comment char', 'Comment Char: '),
            ]
        DictEditDialog.__init__(self, parent, title, "Enter assembler information:", fields, default)

    def can_submit(self):
        control = self.controls['name']
        return len(control.GetValue()) > 0


def prompt_for_assembler(parent, title, default=None):
    d = AssemblerDialog(parent, title, default)
    return d.show_and_get_value()


def get_file_dialog_wildcard(name, extension_list):
    # Using only the first extension
    wildcards = []
    if extension_list:
        ext = extension_list[0]
        wildcards.append("%s (*%s)|*%s" % (name, ext, ext))
    return "|".join(wildcards)


class SegmentOrderDialog(wx.Dialog):
    border = 5
    instructions = "Drag segments to the right-hand list to create the output segment order"
    dest_list_title = "Segments in Executable"

    def __init__(self, parent, title, segments, list_title=None, credits=False):
        wx.Dialog.__init__(self, parent, -1, title, style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)

        self.segment_map = {k:v for k,v in enumerate(segments)}

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        t = wx.StaticText(self, -1, self.instructions)
        sizer.Add(t, 0, wx.ALL|wx.EXPAND, self.border)

        hbox = wx.BoxSizer(wx.HORIZONTAL)

        vbox1 = wx.BoxSizer(wx.VERTICAL)
        t = wx.StaticText(self, -1, "All Segments")
        vbox1.Add(t, 0, wx.ALL|wx.EXPAND, self.border)
        self.source = ReorderableList(self, list(self.segment_map.keys()), self.get_item_text, columns=["Origin", "Size", "Name"], resize_column=3, allow_drop=False, size=(400,300))
        vbox1.Add(self.source, 1, wx.ALL|wx.EXPAND, self.border)
        hbox.Add(vbox1, 1, wx.ALL|wx.EXPAND, 0)

        vbox2 = wx.BoxSizer(wx.VERTICAL)
        if list_title is None:
            list_title = self.dest_list_title
        t = wx.StaticText(self, -1, list_title)
        vbox2.Add(t, 0, wx.ALL|wx.EXPAND, self.border)
        self.dest = ReorderableList(self, [], self.get_item_text, columns=["Origin", "Size", "Name"], resize_column=3, size=(400,300))
        vbox2.Add(self.dest, 1, wx.ALL|wx.EXPAND, self.border)
        hbox.Add(vbox2, 1, wx.ALL|wx.EXPAND, 0)

        vbox = wx.BoxSizer(wx.VERTICAL)

        self.add_command_area(vbox, credits)

        vbox.AddStretchSpacer()

        btnsizer = wx.StdDialogButtonSizer()
        self.ok_btn = wx.Button(self, wx.ID_OK)
        self.ok_btn.SetDefault()
        btnsizer.AddButton(self.ok_btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        vbox.Add(btnsizer, 0, wx.ALL|wx.EXPAND, self.border)
        hbox.Add(vbox, 0, wx.EXPAND, 0)
        sizer.Add(hbox, 1, wx.EXPAND, 0)

        self.Bind(wx.EVT_BUTTON, self.on_button)
        self.Bind(wx.EVT_TEXT, self.on_text_changed)

        # Don't call self.Fit() otherwise the dialog buttons are zero height
        sizer.Fit(self)
        self.check_enable()

    def add_command_area(self, vbox, credits):
        t = wx.StaticText(self, -1, "Run Address")
        vbox.Add(t, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        self.run_addr = wx.TextCtrl(self, -1, size=(-1, -1))
        vbox.Add(self.run_addr, 0, wx.LEFT|wx.RIGHT|wx.BOTTOM|wx.EXPAND, self.border)

        if credits:
            t = wx.StaticText(self, -1, "Title (20 chars)")
            vbox.Add(t, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
            self.title_20 = wx.TextCtrl(self, -1, size=(-1, -1))
            self.title_20.SetMaxLength(20)
            vbox.Add(self.title_20, 0, wx.LEFT|wx.RIGHT|wx.BOTTOM|wx.EXPAND, self.border)

            t = wx.StaticText(self, -1, "Author (20 chars)")
            vbox.Add(t, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
            self.author_20 = wx.TextCtrl(self, -1, size=(-1, -1))
            self.author_20.SetMaxLength(20)
            vbox.Add(self.author_20, 0, wx.LEFT|wx.RIGHT|wx.BOTTOM|wx.EXPAND, self.border)
        else:
            self.title_20 = self.author_20 = None

        vbox.AddSpacer(50)

        self.clear = b = wx.Button(self, wx.ID_DOWN, 'Clear List', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_clear)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)

    def get_item_text(self, sid):
        s = self.segment_map[sid]
        return "%x" % s.origin, "%x" % len(s), s.name

    def on_button(self, evt):
        if evt.GetId() == wx.ID_OK:
            self.EndModal(wx.ID_OK)
        else:
            self.EndModal(wx.ID_CANCEL)
        evt.Skip()

    def on_text_changed(self, evt):
        self.ok_btn.Enable(self.can_submit())

    def on_clear(self, evt):
        self.dest.clear()

    def on_resize(self, event):
        self.Fit()

    def get_run_addr(self):
        text = self.run_addr.GetValue()
        try:
            addr = text_to_int(text, "hex")
            if addr < 0 or addr > 0xffff:
                addr = None
        except (ValueError, TypeError) as e:
            addr = None
        return addr

    def can_submit(self):
        return self.get_run_addr() is not None and self.dest.GetItemCount() > 0

    def check_enable(self):
        self.ok_btn.Enable(self.can_submit())

    def get_segments(self):
        s = []
        for sid in self.dest.items:
            s.append(self.segment_map[sid])
        return s

    def get_document(self):
        segments = self.get_segments()
        root, segs = get_xex(segments)
        doc = DiskImageDocument(bytes=root.raw_bytes, style=root.style)
        Parser = namedtuple("Parser", ['segments'])
        segs[0:0] = [root]
        p = Parser(segments=segs)
        doc.set_segments(p)
        return doc

    def get_extra_text(self):
        lines = []
        if self.title_20 is not None:
            lines.append(self.title_20.GetValue())
        if self.author_20 is not None:
            lines.append(self.author_20.GetValue())
        return lines


class SegmentInterleaveDialog(SegmentOrderDialog):
    instructions = "Drag segments to the right-hand list to determine interleave order. Then, choose interleave factor in bytes."
    dest_list_title = "Interleave Order"

    def add_command_area(self, vbox, credits):
        t = wx.StaticText(self, -1, "Interleave Factor (default hex)")
        t.SetToolTip(wx.ToolTip("Number of bytes from each segment per interleave group"))
        vbox.Add(t, 0, wx.ALL|wx.EXPAND, self.border)
        self.interleave = wx.TextCtrl(self, -1, size=(-1, -1))
        vbox.Add(self.interleave, 0, wx.ALL|wx.EXPAND, self.border)

        vbox.AddSpacer(50)

        self.clear = b = wx.Button(self, wx.ID_DOWN, 'Clear List', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_clear)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)

    def get_length(self):
        segments = self.get_segments()
        if len(segments) == 0:
            return -1
        common = len(segments[0])
        for s in segments[1:]:
            if len(s) != common:
                return -1
        return common

    def get_interleave(self):
        text = self.interleave.GetValue()
        length = self.get_length()
        try:
            num = text_to_int(text, "hex")
            if num < 0 or num > length:
                num = 0
        except (ValueError, TypeError) as e:
            num = 0
        return num

    def can_submit(self):
        return self.get_length() > 0 and self.get_interleave() > 0


if __name__ == "__main__":
    app = wx.PySimpleApp()

    frame = wx.Frame(None, -1, "Dialog test")
#    dialog = EmulatorDialog(frame, "Test")
#    dialog.ShowModal()
#    dlg = ListReorderDialog(frame, [chr(i + 65) for i in range(26)], lambda a:str(a))
    # dlg = CheckItemDialog(frame, [chr(i + 65) for i in range(26)], lambda a:str(a))
    # if dlg.ShowModal() == wx.ID_OK:
    #     print dlg.get_checked_items()
    dlg = ChooseOnePlusCustomDialog(frame, ["one", "two"], "instructions")
    if dlg.ShowModal() == wx.ID_OK:
        print("Selected", dlg.get_selected())
    dlg.Destroy()

    app.MainLoop()



"""Segment actions
"""
import os
import sys

import wx

from sawx.action import SawxAction
from sawx.ui.minibuffer import NextPrevTextMinibuffer

from .. import commands
from .. import errors

import logging
log = logging.getLogger(__name__)


class find(SawxAction):
    def calc_name(self, action_key):
        return "Find"

    def get_minibuffer(self, action_key):
        e = self.editor
        control = NextPrevTextMinibuffer(e, commands.FindAllCommand, commands.FindNextCommand, commands.FindPrevCommand, initial=e.last_search_settings["find"])
        return control

    def perform(self, action_key):
        minibuffer = self.get_minibuffer(action_key)
        self.editor.control.show_minibuffer(minibuffer)


class find_to_selection(SawxAction):
    def calc_name(self, action_key):
        return "Find to Selection"

    def perform(self, action_key):
        e = self.editor
        v = e.focused_viewer
        e.control.on_hide_minibuffer_or_cancel(True)  # refreshes current control, but others need refreshing
        v.segment.convert_style({'match':True}, {'selected':True})
        v.control.caret_handler.convert_style_to_carets()
        flags = e.calc_status_flags()
        flags.refresh_needed = True
        flags.sync_caret_from_control = v.control
        # flags.refreshed_as_side_effect.add(v.control)
        e.process_flags(flags)
